/*
Derived from Examples in 6.2, Network Security with OpenSSL by Pravir Chandra, Matt Messier, John Viega. 
Copyright  2002 O'Reilly & Associates, Inc.

This duplicates the functionality of:
openssl enc -bf-cbc

The format of the data is:
Salted__
salt[8]
data in blocks of 8.
It's very easy.

The default openssl enc -bf-cbc uses md5 hash to generate the key and iv. This 
is bad because md5 has a collision problem. Collisions can be generated 
*/

#include <openssl/evp.h>
#include <openssl/rand.h>
#include <openssl/hmac.h>
#include <openssl/sha.h>

#include <stdio.h>
#include <string.h>

// use a 128-bit key like they say instead of the max key length.
#define EVP_BLOWFISH_KEY_SIZE       16
#define KEY_LENGTH EVP_BLOWFISH_KEY_SIZE
// Header means: blowfish cbc, iv is the first 8 bytes, 
// salt is the next 8 bytes, and the password is sha1 hashed.
#define BF_HEADER "Salted__"

#define MAX_PACKET_SIZE 1024

#ifdef _DEBUG
#define VERBOSE 1
#else
#define VERBOSE 0
#endif

// The idea is that I have 2 programs and one library in one .c file
#ifdef ENCRYPT
#define encrypt_main main
#endif
#ifdef DECRYPT
#define decrypt_main main
#endif

#ifdef _OPENBSD
#define RANDOM_DEVICE "/dev/srandom"
#else
#define RANDOM_DEVICE "/dev/random"
#endif

// If you define this, the key will be generated with SHA1.
// That will make the output incompatible with ones that do not define
// USE_SHA1. It will also make it incompatible with openssl enc -bf-cbc.
// In order to decrypt it with openssl enc -bf-cbc, use the flag: -md sha1
#ifdef USE_SHA1
#define KEY_DIGEST EVP_sha1
#else
#define KEY_DIGEST EVP_md5
#endif

#include <ssl_bf2a.h>

void select_random_key(BYTE *key, int b)
{
    RAND_bytes(key, b);
}

void print_hex(BYTE *data, int b)
{
    int i;

    for (i = 0;  i < b - 1;  i++)
        fprintf(stderr, "%02hhX:", data[i]);
    fprintf(stderr, "%02hhX\n", data[b - 1]);
}

unsigned char *simple_digest(char *buf, unsigned int len, int *olen)
{
    const EVP_MD  *m;
    EVP_MD_CTX    ctx;
    unsigned char *ret;

    char alg[] = "sha1";

    OpenSSL_add_all_digests();
    if (!(m = EVP_get_digestbyname(alg)))
        return NULL;
    if (!(ret = (unsigned char *)malloc(EVP_MAX_MD_SIZE)))
        return NULL;
    EVP_DigestInit(&ctx, m);
    EVP_DigestUpdate(&ctx, buf, len);
    EVP_DigestFinal(&ctx, ret, olen);
    return ret;
}

void get_password(char *pass, int len, int encrypting)
{
    for (;;)
    {
        // If encrypting == 1, verify password.
        pass[0]='\0';
        int i = EVP_read_pw_string((char *)pass, len, "Password:", encrypting);
        if (i == 0)
        {
            if (pass[0] == '\0')
            {
                return;
            }
            //str = pass;
            break;
        }
        if (i < 0)
        {
            fprintf(stderr,"Error: bad password read\n");
            return;
        }
    }
}

void generate_key(BYTE *key, const char *password, const BYTE *salt, BYTE *iv, int b)
{
    const EVP_CIPHER *cipher = NULL;
    cipher = EVP_bf_cbc(); // Error: EVP_get_cipherbyname("bf-cbc");
    const EVP_MD *dgst = NULL;
    dgst = KEY_DIGEST();
    // This is a pretty interesting system.
    // read openssl-xxx/doc/EVP_BytesToKey.pod for interesting info.
    EVP_BytesToKey(cipher, dgst, salt, (BYTE *)password, 
        strlen(password), 1, key, iv);
}

// If the password is a binary instead of a string, put it here.
void generate_bin_key(BYTE *key, BYTE *password, int pass_len, const BYTE *salt, BYTE *iv, int b)
{
    const EVP_CIPHER *cipher = NULL;
    cipher = EVP_bf_cbc(); // Error: EVP_get_cipherbyname("bf-cbc");
    const EVP_MD *dgst = NULL;
    dgst = KEY_DIGEST();
    // This is a pretty interesting system.
    // read openssl-xxx/doc/EVP_BytesToKey.pod for interesting info.
    EVP_BytesToKey(cipher, dgst, salt, password, pass_len, 1, key, iv);
}

void select_random_salt(char *iv, int b)
{
    RAND_pseudo_bytes(iv, b);
}

int seed_prng(int bytes)
{
    // This slows you down if you use it very much.
    if (!RAND_load_file(RANDOM_DEVICE, bytes))
        return 0;
    return 1;
}

char *encrypt_data(EVP_CIPHER_CTX *ctx, char *data, int inl, int *rb)
{
    char *ret;
    int  i, tmp, ol;

    ol = 0;
    // Note: you must free this memory.
    ret = (char *)malloc(inl + EVP_CIPHER_CTX_block_size(ctx));
    // 104 is divisible by 8 which is important.
    for (i = 0;  i < inl / 104;  i++)
    {
        EVP_EncryptUpdate(ctx, &ret[ol], &tmp, &data[ol], 104);
        ol += tmp;
    }
    if (inl % 104)
    {
        EVP_EncryptUpdate(ctx, &ret[ol], &tmp, &data[ol], inl % 104);
        ol += tmp;
    }
    EVP_EncryptFinal(ctx, &ret[ol], &tmp);
    *rb = ol + tmp;
    return ret;
}

char *decrypt_data(EVP_CIPHER_CTX *ctx, char *ct, int inl)
{
    // Note: you must free this memory.
    char *pt = (char *)malloc(inl + EVP_CIPHER_CTX_block_size(ctx));
    int  ol;

    EVP_DecryptUpdate(ctx, pt, &ol, ct, inl);
    if (!ol) /* there's no block to decrypt */
    {
        free(pt);
        return NULL;
    }
    return pt;
}

/* 
The Data Hash check function.
It comes in two parts: generate the HMAC, test the HMAC.
create: Generates a hash to check data by.
check: Returns validation.
Returns 0 on check failure, 1 on success
*/
int HMAC_create(BYTE *data, int dlen, BYTE *key, int klen, BYTE *check, int clen)
{
    BYTE result[SHA_DIGEST_LENGTH];

    int out_len;
    HMAC(EVP_sha1(), key, klen, data, dlen, result, &out_len);
    // Error if HMAC outputs the wrong size.
    if(clen != out_len) return 0;
    // return the HMAC.
    memcpy(check, result, out_len);
    return 1;
}

int HMAC_check(BYTE *data, int dlen, BYTE *key, int klen, BYTE *check, int clen)
{
    BYTE result[SHA_DIGEST_LENGTH];

    int out_len;
    HMAC(EVP_sha1(), key, klen, data, dlen, result, &out_len);
    // Error if HMAC outputs the wrong size.
    if(clen != out_len) return 0;
    // Check it, return 1 if it checks, 0 if it fails.
    return (memcmp(result, check, clen) == 0);
}

/*
The encrypt function:

needs to be called many times in a row.
give it data and a password and you get a packet of data.

The decrypt function:

needs to be called many times in a row.
give it a packet and a password and you get data.

Don't force the developer to do any salt/hmac boloney.
*/
int encrypt_packet(BYTE *data, int data_size, char *password, BYTE *out, int *enc_size)
{
    EVP_CIPHER_CTX ctx;
    BYTE           key[KEY_LENGTH]; 
    BYTE           iv[EVP_MAX_IV_LENGTH];
    BYTE           salt[PKCS5_SALT_LEN];
    BYTE           hmac[SHA_DIGEST_LENGTH];
    BYTE           hkey[KEY_LENGTH]; 

    // PKCS5_SALT_LEN will always be 8.
    select_random_salt(salt, PKCS5_SALT_LEN);
    
    generate_key(key, password, salt, iv, KEY_LENGTH);
    //print_hex(key, KEY_LENGTH);
    if(key == 0)
    {
        fprintf(stderr, "Error: Key generation failed.\n");
        return 1;
    }
    // iv is selected by the password+salt.

    EVP_EncryptInit(&ctx, EVP_bf_cbc(), key, iv);
    char *enc_data = encrypt_data(&ctx, data, data_size, enc_size);
    // Do not use the same key for encryption and MACing under any circumstances.
    {
        // hiv is not used, so make it a local variable.
        BYTE hiv[EVP_MAX_IV_LENGTH];
        // Generate the hash key by hashing the key and the salt.
        // Might not be great, but it's better than nothing 
        // and it's fairly secure.
        // Should I encrypt the HMAC? No, not necessary.
        memset(hkey, 0, KEY_LENGTH);
        generate_bin_key(hkey, key, KEY_LENGTH, salt, hiv, KEY_LENGTH);
        //print_hex(hkey, KEY_LENGTH);
    }
    if(HMAC_create(data, data_size, hkey, KEY_LENGTH, hmac, SHA_DIGEST_LENGTH) == 0)
    {
        fprintf(stderr, "Error: HMAC generation failed.\n");
        return 1;
    }
    //printf("Ciphertext is %d bytes.\n", enc_size);
    memcpy(out, BF_HEADER, 8);
    memcpy(out+8, salt, PKCS5_SALT_LEN);
    memcpy(out+8+PKCS5_SALT_LEN, hmac, SHA_DIGEST_LENGTH);
    memcpy(out+8+PKCS5_SALT_LEN+SHA_DIGEST_LENGTH, enc_data, *enc_size);
    //fprintf(stderr, "enc_size = %i + 8 + %i\n", *enc_size, PKCS5_SALT_LEN);
    // Add the header length and the salt length and the hmac length.
    *enc_size += 8 + PKCS5_SALT_LEN + SHA_DIGEST_LENGTH;
    // FIXME: Memory leak. *** glibc detected *** double free or corruption (!prev): 0x0804d1c0 ***
    //free(enc_data);
    return 0;
}

int decrypt_packet(BYTE *enc_data, int enc_size, char *password, BYTE *out, int *data_size)
{
    EVP_CIPHER_CTX ctx;
    BYTE           key[KEY_LENGTH]; 
    BYTE           iv[EVP_MAX_IV_LENGTH];
    BYTE           salt[PKCS5_SALT_LEN];
    BYTE           hmac[SHA_DIGEST_LENGTH];
    BYTE           hkey[KEY_LENGTH]; 

    char *dec_data;
    char *ptr_enc;
    char  final[EVP_MAX_BLOCK_LENGTH];
    int   ret = 0;

    // Take the encrypted data from the stdin.
    if(0 != strncmp(enc_data, BF_HEADER, 8))
    {
        fprintf(stderr, "Error: Header not found.\n");
        return 1;
    }
    memcpy(salt, enc_data + 8, PKCS5_SALT_LEN);
    memcpy(hmac, enc_data + 8 + PKCS5_SALT_LEN, SHA_DIGEST_LENGTH);
    
    generate_key(key, password, salt, iv, KEY_LENGTH);
    //print_hex(key, KEY_LENGTH);
    ptr_enc = enc_data + 8 + PKCS5_SALT_LEN + SHA_DIGEST_LENGTH;
    enc_size = enc_size - (8 + PKCS5_SALT_LEN + SHA_DIGEST_LENGTH);
    
    EVP_DecryptInit(&ctx, EVP_bf_cbc(), key, iv);
    int j, k = 0;
    for(j=0; j<enc_size; j+=8)
    {
        dec_data = decrypt_data(&ctx, ptr_enc + j, 8);
        // The first one is always null, ignore it or fix it.
        // TODO: You should probably try to understand it as well.
        if(dec_data != 0)
        {
            memcpy(out + k, dec_data, 8);
            free(dec_data);
            // Counter for output.
            k += 8;
        }
        //else
        //    fprintf(stderr, "Error: Decrypt failed, null data.\n");
    }
    if (!EVP_DecryptFinal(&ctx, final, data_size))
    {
        fprintf(stderr, "Error: Decrypt Final, padding incorrect.\n");
        ret = 1;
        //abort();
    }
    memcpy(out + k, final, *data_size);
    // Data size is set in DecryptFinal to the last size (0-7).
    // Add k (output counter) to get the correct total size.
    *data_size += k;
    // Do not use the same key for encryption and MACing under any circumstances.
    {
        // hiv is not used, so make it a local variable.
        BYTE hiv[EVP_MAX_IV_LENGTH];
        // Generate the hash key by hashing the key and the salt.
        // Might not be great, but it's better than nothing 
        // and it's fairly secure.
        // Should I encrypt the HMAC? No, not necessary.
        memset(hkey, 0, KEY_LENGTH);
        generate_bin_key(hkey, key, KEY_LENGTH, salt, hiv, KEY_LENGTH);
        //print_hex(hkey, KEY_LENGTH);
    }
    if(HMAC_check(out, *data_size, hkey, KEY_LENGTH, hmac, SHA_DIGEST_LENGTH) != 1)
    {
        // Send an error if the hash fails.
        fprintf(stderr, "Error: HMAC hash check failed.\n");
        ret = 1;
    }
    return ret;
}

int encrypt_main(int argc, char *argv[])
{
    if (!seed_prng(4))
    {
        fprintf(stderr, "Error: Unable to seed the PRNG.\n");
        abort();
    }

    char           enc_data[MAX_PACKET_SIZE + 52];
    char           data[MAX_PACKET_SIZE];
    char           password[256];
    
    // Secure password retreival algorithm from enc
    get_password(password, 256, 1);
    
    // Read binary data up to 1024 bytes from the stdin.
    int data_size = fread(data, 1, MAX_PACKET_SIZE, stdin);
    //data[data_size] = 0;
    //fprintf(stderr, "data: '%i'\n", data_size);

    int enc_size = -1;
    int ret = encrypt_packet(data, data_size, password, enc_data, &enc_size);
    
    fwrite(enc_data, enc_size, 1, stdout);
    
    return ret;
}

int decrypt_main(int argc, char *argv[])
{
    if (!seed_prng(4))
    {
        fprintf(stderr, "Error: Unable to seed the PRNG.\n");
        abort();
    }

    char           enc_data[MAX_PACKET_SIZE + 52];
    char           data[MAX_PACKET_SIZE];
    char           password[256];
    
    // Secure password retreival algorithm from enc
    get_password(password, 256, 0);
    
    // Read binary data up to 1024 bytes from the stdin.
    int enc_size = fread(enc_data, 1, MAX_PACKET_SIZE + 24 + PKCS5_SALT_LEN + SHA_DIGEST_LENGTH, stdin);
    int data_size = -1;
    int ret = decrypt_packet(enc_data, enc_size, password, data, &data_size);
    
    fwrite(data, data_size, 1, stdout);
    
    return ret;
}



